import json
import logging
import re
import threading

from plain.http import ResponseRedirect
from plain.json import PlainJSONEncoder
from plain.models import connection
from plain.runtime import settings
from plain.urls import reverse

from .core import QueryStats

try:
    try:
        import psycopg
    except ImportError:
        import psycopg2 as psycopg
except ImportError:
    psycopg = None

logger = logging.getLogger(__name__)
_local = threading.local()


class QueryStatsJSONEncoder(PlainJSONEncoder):
    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            if psycopg and isinstance(obj, psycopg._json.Json):
                return obj.adapted
            else:
                raise


class QueryStatsMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.ignore_url_patterns = [
            re.compile(url) for url in settings.QUERYSTATS_IGNORE_URLS
        ]

    def should_ignore_request(self, request):
        for url in self.ignore_url_patterns:
            if url.match(request.path):
                return True

        return False

    def __call__(self, request):
        if request.GET.get("querystats") == "disable":
            return self.get_response(request)

        querystats = QueryStats(
            # Only want these if we're getting ready to show it
            include_tracebacks=request.GET.get("querystats") == "store"
        )

        with connection.execute_wrapper(querystats):
            # Have to wrap this first call so it is included in the querystats,
            # but we don't have to wrap everything else unless we are staff or debug
            is_staff = self.is_staff_request(request)

        if (settings.DEBUG or is_staff) and not self.should_ignore_request(request):
            # Persist it on the thread
            _local.querystats = querystats

            with connection.execute_wrapper(_local.querystats):
                response = self.get_response(request)

            if settings.DEBUG:
                # TODO logging settings
                logger.debug("Querystats: %s", _local.querystats)

            # Make current querystats available on the current page
            # by using the server timing API which can be parsed client-side
            response.headers["Server-Timing"] = _local.querystats.as_server_timing()

            if request.GET.get("querystats") == "store":
                request.session["querystats"] = json.dumps(
                    _local.querystats.as_context_dict(), cls=QueryStatsJSONEncoder
                )
                return ResponseRedirect(reverse("querystats:querystats"))

            del _local.querystats

            return response

        else:
            return self.get_response(request)

    @staticmethod
    def is_staff_request(request):
        if getattr(request, "impersonator", None):
            # Support for impersonation (still want the real staff user to see the querystats)
            return request.impersonator and request.impersonator.is_staff

        return hasattr(request, "user") and request.user and request.user.is_staff



SECRET_KEY = "test"
INSTALLED_PACKAGES = [
    "plain.auth",
    "plain.sessions",
    "plain.models",
    "plain.htmx",
    "plain.tailwind",
    "plain.staff",
    "app.users",
]
DATABASES = {
    "default": {
        "ENGINE": "plain.models.backends.sqlite3",
        "NAME": ":memory:",
    }
}
MIDDLEWARE = [
    "plain.sessions.middleware.SessionMiddleware",
    "plain.auth.middleware.AuthenticationMiddleware",
    "plain.staff.StaffMiddleware",
]
AUTH_LOGIN_URL = "login"
AUTH_USER_MODEL = "users.User"



import datetime
from calendar import monthrange
from enum import Enum

from plain.utils import timezone


class DatetimeRangeAliases(Enum):
    TODAY = "Today"
    THIS_WEEK = "This Week"
    THIS_WEEK_TO_DATE = "This Week-to-date"
    THIS_MONTH = "This Month"
    THIS_MONTH_TO_DATE = "This Month-to-date"
    THIS_QUARTER = "This Quarter"
    THIS_QUARTER_TO_DATE = "This Quarter-to-date"
    THIS_YEAR = "This Year"
    THIS_YEAR_TO_DATE = "This Year-to-date"
    LAST_WEEK = "Last Week"
    LAST_WEEK_TO_DATE = "Last Week-to-date"
    LAST_MONTH = "Last Month"
    LAST_MONTH_TO_DATE = "Last Month-to-date"
    LAST_QUARTER = "Last Quarter"
    LAST_QUARTER_TO_DATE = "Last Quarter-to-date"
    LAST_YEAR = "Last Year"
    LAST_YEAR_TO_DATE = "Last Year-to-date"
    SINCE_30_DAYS_AGO = "Since 30 Days Ago"
    SINCE_60_DAYS_AGO = "Since 60 Days Ago"
    SINCE_90_DAYS_AGO = "Since 90 Days Ago"
    SINCE_365_DAYS_AGO = "Since 365 Days Ago"
    NEXT_WEEK = "Next Week"
    NEXT_4_WEEKS = "Next 4 Weeks"
    NEXT_MONTH = "Next Month"
    NEXT_QUARTER = "Next Quarter"
    NEXT_YEAR = "Next Year"

    # TODO doesn't include anything less than a day...
    # ex. SINCE_1_HOUR_AGO = "Since 1 Hour Ago"

    def __str__(self):
        return self.value

    @classmethod
    def from_value(cls, value):
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"{value} is not a valid value for {cls.__name__}")

    @classmethod
    def to_range(cls, value: str) -> tuple[datetime.datetime, datetime.datetime]:
        now = timezone.localtime()
        start_of_today = now.replace(hour=0, minute=0, second=0, microsecond=0)
        start_of_week = start_of_today - datetime.timedelta(
            days=start_of_today.weekday()
        )
        start_of_month = start_of_today.replace(day=1)
        start_of_quarter = start_of_today.replace(
            month=((start_of_today.month - 1) // 3) * 3 + 1, day=1
        )
        start_of_year = start_of_today.replace(month=1, day=1)

        def end_of_day(dt):
            return dt.replace(hour=23, minute=59, second=59, microsecond=999999)

        def end_of_month(dt):
            last_day = monthrange(dt.year, dt.month)[1]
            return end_of_day(dt.replace(day=last_day))

        def end_of_quarter(dt):
            end_month = ((dt.month - 1) // 3 + 1) * 3
            return end_of_month(dt.replace(month=end_month))

        def end_of_year(dt):
            return end_of_month(dt.replace(month=12))

        if value == cls.TODAY:
            return DatetimeRange(start_of_today, end_of_day(now))
        if value == cls.THIS_WEEK:
            return DatetimeRange(
                start_of_week, end_of_day(start_of_week + datetime.timedelta(days=6))
            )
        if value == cls.THIS_WEEK_TO_DATE:
            return DatetimeRange(start_of_week, now)
        if value == cls.THIS_MONTH:
            return DatetimeRange(start_of_month, end_of_month(start_of_month))
        if value == cls.THIS_MONTH_TO_DATE:
            return DatetimeRange(start_of_month, now)
        if value == cls.THIS_QUARTER:
            return DatetimeRange(start_of_quarter, end_of_quarter(start_of_quarter))
        if value == cls.THIS_QUARTER_TO_DATE:
            return DatetimeRange(start_of_quarter, now)
        if value == cls.THIS_YEAR:
            return DatetimeRange(start_of_year, end_of_year(start_of_year))
        if value == cls.THIS_YEAR_TO_DATE:
            return DatetimeRange(start_of_year, now)
        if value == cls.LAST_WEEK:
            last_week_start = start_of_week - datetime.timedelta(days=7)
            return DatetimeRange(
                last_week_start,
                end_of_day(last_week_start + datetime.timedelta(days=6)),
            )
        if value == cls.LAST_WEEK_TO_DATE:
            return DatetimeRange(start_of_week - datetime.timedelta(days=7), now)
        if value == cls.LAST_MONTH:
            last_month = (start_of_month - datetime.timedelta(days=1)).replace(day=1)
            return DatetimeRange(last_month, end_of_month(last_month))
        if value == cls.LAST_MONTH_TO_DATE:
            last_month = (start_of_month - datetime.timedelta(days=1)).replace(day=1)
            return DatetimeRange(last_month, now)
        if value == cls.LAST_QUARTER:
            last_quarter = (start_of_quarter - datetime.timedelta(days=1)).replace(
                day=1
            )
            return DatetimeRange(last_quarter, end_of_quarter(last_quarter))
        if value == cls.LAST_QUARTER_TO_DATE:
            last_quarter = (start_of_quarter - datetime.timedelta(days=1)).replace(
                day=1
            )
            return DatetimeRange(last_quarter, now)
        if value == cls.LAST_YEAR:
            last_year = start_of_year.replace(year=start_of_year.year - 1)
            return DatetimeRange(last_year, end_of_year(last_year))
        if value == cls.LAST_YEAR_TO_DATE:
            last_year = start_of_year.replace(year=start_of_year.year - 1)
            return DatetimeRange(last_year, now)
        if value == cls.SINCE_30_DAYS_AGO:
            return DatetimeRange(now - datetime.timedelta(days=30), now)
        if value == cls.SINCE_60_DAYS_AGO:
            return DatetimeRange(now - datetime.timedelta(days=60), now)
        if value == cls.SINCE_90_DAYS_AGO:
            return DatetimeRange(now - datetime.timedelta(days=90), now)
        if value == cls.SINCE_365_DAYS_AGO:
            return DatetimeRange(now - datetime.timedelta(days=365), now)
        if value == cls.NEXT_WEEK:
            next_week_start = start_of_week + datetime.timedelta(days=7)
            return DatetimeRange(
                next_week_start,
                end_of_day(next_week_start + datetime.timedelta(days=6)),
            )
        if value == cls.NEXT_4_WEEKS:
            return DatetimeRange(now, end_of_day(now + datetime.timedelta(days=28)))
        if value == cls.NEXT_MONTH:
            next_month = (start_of_month + datetime.timedelta(days=31)).replace(day=1)
            return DatetimeRange(next_month, end_of_month(next_month))
        if value == cls.NEXT_QUARTER:
            next_quarter = (start_of_quarter + datetime.timedelta(days=90)).replace(
                day=1
            )
            return DatetimeRange(next_quarter, end_of_quarter(next_quarter))
        if value == cls.NEXT_YEAR:
            next_year = start_of_year.replace(year=start_of_year.year + 1)
            return DatetimeRange(next_year, end_of_year(next_year))
        raise ValueError(f"Invalid range: {value}")


class DatetimeRange:
    def __init__(self, start, end):
        self.start = start
        self.end = end

        if isinstance(self.start, str) and self.start:
            self.start = datetime.datetime.fromisoformat(self.start)

        if isinstance(self.end, str) and self.end:
            self.end = datetime.datetime.fromisoformat(self.end)

        if isinstance(self.start, datetime.date) and not isinstance(
            self.start, datetime.datetime
        ):
            self.start = timezone.localtime().replace(
                year=self.start.year, month=self.start.month, day=self.start.day
            )

        if isinstance(self.end, datetime.date) and not isinstance(
            self.start, datetime.datetime
        ):
            self.end = timezone.localtime().replace(
                year=self.end.year, month=self.end.month, day=self.end.day
            )

    def as_tuple(self):
        return (self.start, self.end)

    def total_days(self):
        return (self.end - self.start).days

    def iter_days(self):
        """Yields each day in the range."""
        return iter(
            self.start.date() + datetime.timedelta(days=i)
            for i in range(0, self.total_days())
        )

    def iter_weeks(self):
        """Yields the start of each week in the range."""
        current = self.start - datetime.timedelta(days=self.start.weekday())
        current = current.replace(hour=0, minute=0, second=0, microsecond=0)
        while current <= self.end:
            next_week = current + datetime.timedelta(weeks=1)
            yield current
            current = next_week

    def iter_months(self):
        """Yields the start of each month in the range."""
        current = self.start.replace(day=1)
        current = current.replace(hour=0, minute=0, second=0, microsecond=0)
        while current <= self.end:
            if current.month == 12:
                next_month = current.replace(year=current.year + 1, month=1)
            else:
                next_month = current.replace(month=current.month + 1)
            yield current
            current = next_month

    def iter_quarters(self):
        """Yields the start of each quarter in the range."""
        current = self.start.replace(month=((self.start.month - 1) // 3) * 3 + 1, day=1)
        current = current.replace(hour=0, minute=0, second=0, microsecond=0)
        while current <= self.end:
            next_quarter_month = ((current.month - 1) // 3 + 1) * 3 + 1
            if next_quarter_month > 12:
                next_quarter_month -= 12
                next_year = current.year + 1
            else:
                next_year = current.year
            next_quarter = datetime.datetime(
                next_year, next_quarter_month, 1, tzinfo=current.tzinfo
            )
            yield current
            current = next_quarter

    def iter_years(self):
        """Yields the start of each year in the range."""
        current = self.start.replace(month=1, day=1)
        current = current.replace(hour=0, minute=0, second=0, microsecond=0)
        while current <= self.end:
            next_year = current.replace(year=current.year + 1)
            yield current
            current = next_year

    def __repr__(self):
        return f"DatetimeRange({self.start}, {self.end})"

    def __str__(self):
        return f"{self.start} to {self.end}"

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end

    def __hash__(self):
        return hash((self.start, self.end))

    def __contains__(self, item):
        return self.start <= item <= self.end



from plain.http import ResponseRedirect
from plain.urls import include, path

from .impersonate import urls as impersonate_urls
from .querystats import urls as querystats_urls
from .views.base import StaffView
from .views.registry import registry


class StaffIndexView(StaffView):
    template_name = "staff/index.html"
    title = "Dashboard"
    slug = ""

    def get(self):
        # Slight hack to redirect to the first view that doesn't
        # require any url params...
        if views := registry.get_searchable_views():
            return ResponseRedirect(list(views)[0].get_absolute_url())

        return super().get()


class StaffSearchView(StaffView):
    template_name = "staff/search.html"
    title = "Search"
    slug = "search"

    def get_template_context(self):
        context = super().get_template_context()
        context["searchable_views"] = registry.get_searchable_views()
        context["global_search_query"] = self.request.GET.get("query", "")
        return context


default_namespace = "staff"


urlpatterns = [
    path("search/", StaffSearchView, name="search"),
    path("impersonate/", include(impersonate_urls)),
    path("querystats/", include(querystats_urls)),
    path("", include(registry.get_urls())),
    path("", StaffIndexView, name="index"),
]



from plain.auth import get_user_model
from plain.http import ResponseForbidden

from .permissions import can_be_impersonator, can_impersonate_user
from .views import IMPERSONATE_KEY


def get_user_by_pk(pk):
    UserModel = get_user_model()

    try:
        return UserModel.objects.get(pk=pk)
    except UserModel.DoesNotExist:
        return None


class ImpersonateMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if (
            IMPERSONATE_KEY in request.session
            and request.user
            and can_be_impersonator(request.user)
        ):
            user_to_impersonate = get_user_by_pk(request.session[IMPERSONATE_KEY])
            if user_to_impersonate:
                if not can_impersonate_user(request.user, user_to_impersonate):
                    # Can't impersonate this user, remove it and show an error
                    del request.session[IMPERSONATE_KEY]
                    return ResponseForbidden()

                # Finally, change the request user and keep a reference to the original
                request.impersonator = request.user
                request.user = user_to_impersonate

        return self.get_response(request)



from .base import (
    StaffCreateView,
    StaffDeleteView,
    StaffDetailView,
    StaffListView,
    StaffUpdateView,
    StaffView,
)
from .models import (
    StaffModelCreateView,
    StaffModelDeleteView,
    StaffModelDetailView,
    StaffModelListView,
    StaffModelUpdateView,
    StaffModelViewset,
)
from .registry import (
    get_model_detail_url,
    register_view,
    register_viewset,
)
from .types import Img

__all__ = [
    "StaffView",
    "StaffListView",
    "StaffCreateView",
    "StaffUpdateView",
    "StaffDetailView",
    "StaffDeleteView",
    "StaffModelViewset",
    "StaffModelListView",
    "StaffModelCreateView",
    "StaffModelDetailView",
    "StaffModelUpdateView",
    "StaffModelDeleteView",
    "register_viewset",
    "register_view",
    "get_model_detail_url",
    "Img",
]



from collections import defaultdict

from plain.models import Count
from plain.models.functions import (
    TruncDate,
    TruncMonth,
)
from plain.staff.dates import DatetimeRangeAliases

from .base import Card


class ChartCard(Card):
    template_name = "staff/cards/chart.html"

    def get_template_context(self):
        context = super().get_template_context()
        context["chart_data"] = self.get_chart_data()
        return context

    def get_chart_data(self) -> dict:
        raise NotImplementedError


class TrendCard(ChartCard):
    """
    A card that renders a trend chart.
    Primarily intended for use with models, but it can also be customized.
    """

    model = None
    datetime_field = None
    default_display = DatetimeRangeAliases.SINCE_30_DAYS_AGO

    displays = DatetimeRangeAliases

    def get_description(self):
        datetime_range = DatetimeRangeAliases.to_range(self.get_current_display())
        return f"{datetime_range.start} to {datetime_range.end}"

    def get_current_display(self):
        if s := super().get_current_display():
            return DatetimeRangeAliases.from_value(s)
        return self.default_display

    def get_trend_data(self) -> list[int | float]:
        if not self.model or not self.datetime_field:
            raise NotImplementedError(
                "model and datetime_field must be set, or get_values must be overridden"
            )

        datetime_range = DatetimeRangeAliases.to_range(self.get_current_display())

        filter_kwargs = {f"{self.datetime_field}__range": datetime_range.as_tuple()}

        if datetime_range.total_days() < 300:
            truncator = TruncDate
            iterator = datetime_range.iter_days
        else:
            truncator = TruncMonth
            iterator = datetime_range.iter_months

        counts_by_date = (
            self.model.objects.filter(**filter_kwargs)
            .annotate(chart_date=truncator(self.datetime_field))
            .values("chart_date")
            .annotate(chart_date_count=Count("id"))
        )

        # Will do the zero filling for us on key access
        date_values = defaultdict(int)

        for row in counts_by_date:
            date_values[row["chart_date"]] = row["chart_date_count"]

        return {date.strftime("%Y-%m-%d"): date_values[date] for date in iterator()}

    def get_chart_data(self) -> dict:
        data = self.get_trend_data()
        trend_labels = list(data.keys())
        trend_data = list(data.values())

        def calculate_trend_line(data):
            """
            Calculate a trend line using basic linear regression.
            :param data: A list of numeric values representing the y-axis.
            :return: A list of trend line values (same length as data).
            """
            if not data or len(data) < 2:
                return (
                    data  # Return the data as-is if not enough points for a trend line
                )

            n = len(data)
            x = list(range(n))
            y = data

            # Calculate the means of x and y
            x_mean = sum(x) / n
            y_mean = sum(y) / n

            # Calculate the slope (m) and y-intercept (b) of the line: y = mx + b
            numerator = sum((x[i] - x_mean) * (y[i] - y_mean) for i in range(n))
            denominator = sum((x[i] - x_mean) ** 2 for i in range(n))
            slope = numerator / denominator if denominator != 0 else 0
            intercept = y_mean - slope * x_mean

            # Calculate the trend line values
            trend = [slope * xi + intercept for xi in x]

            # if it's all zeros, return nothing
            if all(v == 0 for v in trend):
                return []

            return trend

        return {
            "type": "bar",
            "data": {
                "labels": trend_labels,
                "datasets": [
                    {
                        "data": trend_data,
                    },
                    {
                        "data": calculate_trend_line(trend_data),
                        "type": "line",
                        "borderColor": "rgba(0, 0, 0, 0.3)",
                        "borderWidth": 2,
                        "fill": False,
                        "pointRadius": 0,  # Optional: Hide points
                    },
                ],
            },
            # Hide the label
            # "options": {"legend": {"display": False}},
            # Hide the scales
            "options": {
                "plugins": {"legend": {"display": False}},
                "scales": {
                    "x": {
                        "display": False,
                    },
                    "y": {
                        "suggestedMin": 0,
                    },
                },
                "maintainAspectRatio": False,
                "elements": {
                    "bar": {"borderRadius": "3", "backgroundColor": "rgb(28, 25, 23)"}
                },
            },
        }



from plain.http import ResponseForbidden, ResponseRedirect
from plain.views import View

from .permissions import can_be_impersonator

IMPERSONATE_KEY = "impersonate"


class ImpersonateStartView(View):
    def get(self):
        # We *could* already be impersonating, so need to consider that
        impersonator = getattr(self.request, "impersonator", self.request.user)
        if impersonator and can_be_impersonator(impersonator):
            self.request.session[IMPERSONATE_KEY] = self.url_kwargs["pk"]
            return ResponseRedirect(self.request.GET.get("next", "/"))

        return ResponseForbidden()


class ImpersonateStopView(View):
    def get(self):
        self.request.session.pop(IMPERSONATE_KEY)
        return ResponseRedirect(self.request.GET.get("next", "/"))



from plain.urls import path, reverse_lazy


class StaffViewRegistry:
    def __init__(self):
        # View classes that will be added to the admin automatically
        self.registered_views = set()

    def register_view(self, view=None):
        def inner(view):
            self.registered_views.add(view)
            # TODO do this somewhere else...
            # self.registered_views = set(self.registered_views, key=lambda v: v.title)
            return view

        if callable(view):
            return inner(view)
        else:
            return inner

    def register_viewset(self, viewset=None):
        def inner(viewset):
            for view in viewset.get_views():
                self.register_view(view)
            return viewset

        if callable(viewset):
            return inner(viewset)
        else:
            return inner

    def get_nav_sections(self):
        # class NavItem:
        #     def __init__(self, view, children):
        #         self.view = view
        #         self.children = children

        sections = {}

        for view in self.registered_views:
            section = view.get_nav_section()
            if not section:
                continue
            if section not in sections:
                sections[section] = []
            sections[section].append(view)

        # Sort each section by nav_title
        for section in sections.values():
            section.sort(key=lambda v: v.get_nav_title())

        # Sort sections dictionary by key
        sections = dict(sorted(sections.items()))

        return sections

        # root_nav_items = []

        # for view in sorted_views:
        #     if view.parent_view_class:
        #         continue
        #     children = [x for x in sorted_views if x.parent_view_class == view]
        #     root_nav_items.append(NavItem(view, children))

        # return root_nav_items

    def get_urls(self):
        urlpatterns = []

        paths_seen = set()

        def add_view_path(view, _path):
            if _path in paths_seen:
                raise ValueError(f"Path {_path} already registered")
            paths_seen.add(_path)
            if not _path.endswith("/"):
                _path += "/"
            urlpatterns.append(path(_path, view, name=view.view_name()))

        for view in self.registered_views:
            add_view_path(view, f"p/{view.get_path()}")

        return urlpatterns

    def get_searchable_views(self):
        views = [
            view
            for view in self.registered_views
            if getattr(view, "allow_global_search", False)
        ]
        # Sort by slug since title isn't required by all views
        views.sort(key=lambda v: v.get_slug())
        return views

    def get_model_detail_url(self, instance):
        from plain.staff.views.base import URL_NAMESPACE
        from plain.staff.views.models import StaffModelDetailView

        if not instance.pk:
            # Has to actually be in the db
            return

        for view in self.registered_views:
            if not issubclass(view, StaffModelDetailView):
                continue

            if view.model == instance.__class__:
                return reverse_lazy(
                    f"{URL_NAMESPACE}:{view.view_name()}",
                    kwargs={"pk": instance.pk},
                )


registry = StaffViewRegistry()
register_view = registry.register_view
register_viewset = registry.register_viewset
get_model_detail_url = registry.get_model_detail_url



import json

from plain.auth.views import AuthViewMixin
from plain.views import TemplateView


class QuerystatsView(AuthViewMixin, TemplateView):
    template_name = "querystats/querystats.html"
    staff_required = True  # allow impersonator?

    def get_template_context(self):
        context = super().get_template_context()

        stored_querystats = self.request.session.get(
            "querystats"
        )  # Not popping so page can be reloaded
        if stored_querystats:
            # dates won't come back as Python dates...
            stored_querystats = json.loads(stored_querystats)
            context["querystats"] = stored_querystats

        return context

    def get_querystats(self):
        from .middleware import _local

        return _local.querystats



from importlib import import_module
from importlib.util import find_spec

from plain.packages import PackageConfig, packages

MODULE_NAME = "staff"


class Config(PackageConfig):
    name = "plain.staff"
    label = "plainstaff"

    def ready(self):
        def _import_if_exists(module_name):
            if find_spec(module_name):
                import_module(module_name)

        # Trigger register calls to fire by importing the modules
        for package_config in packages.get_package_configs():
            _import_if_exists(f"{package_config.name}.{MODULE_NAME}")

        # Also trigger for the root app/staff.py module
        _import_if_exists(f"app.{MODULE_NAME}")



from typing import TYPE_CHECKING

from plain.auth.views import AuthViewMixin
from plain.htmx.views import HTMXViewMixin
from plain.http import Response, ResponseRedirect
from plain.paginator import Paginator
from plain.urls import reverse
from plain.utils import timezone
from plain.utils.text import slugify
from plain.views import (
    CreateView,
    DeleteView,
    DetailView,
    TemplateView,
    UpdateView,
)

from .registry import registry
from .types import Img

if TYPE_CHECKING:
    from ..cards import Card


URL_NAMESPACE = "staff"


class StaffView(AuthViewMixin, TemplateView):
    staff_required = True

    title: str = ""
    slug: str = ""
    path: str = ""
    description: str = ""
    image: Img | None = None

    # Leave empty to hide from nav
    #
    # An explicit disabling of showing this url/page in the nav
    # which importantly effects the (future) recent pages list
    # so you can also use this for pages that can never be bookmarked
    nav_section = "App"
    nav_title = ""

    links: dict[str] = {}

    parent_view_class: "StaffView" = None

    template_name = "staff/page.html"
    cards: list["Card"] = []

    def get_template_context(self):
        context = super().get_template_context()
        context["title"] = self.get_title()
        context["image"] = self.get_image()
        context["slug"] = self.get_slug()
        context["description"] = self.get_description()
        context["links"] = self.get_links()
        context["parent_view_classes"] = self.get_parent_view_classes()
        context["admin_registry"] = registry
        context["cards"] = self.get_cards()
        context["render_card"] = lambda card: card().render(self, self.request)
        context["time_zone"] = timezone.get_current_timezone_name()
        return context

    @classmethod
    def view_name(cls) -> str:
        return f"view_{cls.get_slug()}"

    @classmethod
    def get_slug(cls) -> str:
        if cls.slug:
            return cls.slug

        if cls.title:
            return slugify(cls.title)

        raise NotImplementedError(
            f"Please set a slug on the {cls} class or implement get_slug()."
        )

    # Can actually use @classmethod, @staticmethod or regular method for these?
    def get_title(self) -> str:
        return self.title

    def get_image(self) -> Img | None:
        return self.image

    def get_description(self) -> str:
        return self.description

    @classmethod
    def get_path(cls) -> str:
        if cls.path:
            return cls.path

        if slug := cls.get_slug():
            return slug

        raise NotImplementedError(
            f"Please set a path on the {cls} class or implement get_slug() or get_path()."
        )

    @classmethod
    def get_parent_view_classes(cls) -> list["StaffView"]:
        parents = []
        parent = cls.parent_view_class
        while parent:
            parents.append(parent)
            parent = parent.parent_view_class
        return parents

    @classmethod
    def get_nav_section(cls) -> bool:
        if not cls.nav_section:
            return ""

        if cls.parent_view_class:
            # Don't show child views by default
            return ""

        return cls.nav_section

    @classmethod
    def get_nav_title(cls) -> str:
        if cls.nav_title:
            return cls.nav_title

        if cls.title:
            return cls.title

        raise NotImplementedError(
            f"Please set a title or nav_title on the {cls} class or implement get_nav_title()."
        )

    @classmethod
    def get_absolute_url(cls) -> str:
        return reverse(f"{URL_NAMESPACE}:" + cls.view_name())

    def get_links(self) -> dict[str]:
        return self.links.copy()

    def get_cards(self):
        return self.cards.copy()


class StaffListView(HTMXViewMixin, StaffView):
    template_name = "staff/list.html"
    fields: list[str]
    actions: list[str] = []
    displays: list[str] = []
    page_size = 100
    show_search = False
    allow_global_search = False

    def get_template_context(self):
        context = super().get_template_context()

        # Make this available on self for usage in get_objects and other methods
        self.display = self.request.GET.get("display", "")

        # Make this available to get_displays and stuff
        self.objects = self.get_objects()

        page_size = self.request.GET.get("page_size", self.page_size)
        paginator = Paginator(self.objects, page_size)
        self._page = paginator.get_page(self.request.GET.get("page", 1))

        context["paginator"] = paginator
        context["page"] = self._page
        context["objects"] = self._page  # alias
        context["fields"] = self.get_fields()
        context["actions"] = self.get_actions()
        context["displays"] = self.get_displays()

        context["current_display"] = self.display

        # Implement search yourself in get_objects
        context["search_query"] = self.request.GET.get("search", "")
        context["show_search"] = self.show_search

        context["table_style"] = getattr(self, "_table_style", "default")

        context["get_object_pk"] = self.get_object_pk
        context["get_field_value"] = self.get_field_value
        context["get_field_value_template"] = self.get_field_value_template

        context["get_object_url"] = self.get_object_url
        context["get_object_links"] = self.get_object_links

        return context

    def get(self) -> Response:
        if self.is_htmx_request:
            hx_from_this_page = self.request.path in self.request.headers.get(
                "HX-Current-Url", ""
            )
            if not hx_from_this_page:
                self._table_style = "simple"
        else:
            hx_from_this_page = False

        response = super().get()

        if self.is_htmx_request and not hx_from_this_page and not self._page:
            # Don't render anything
            return Response(status=204)

        return response

    def post(self) -> Response:
        # won't be "key" anymore, just list
        action_name = self.request.POST.get("action_name")
        actions = self.get_actions()
        if action_name and action_name in actions:
            target_pks = self.request.POST["action_pks"].split(",")
            response = self.perform_action(action_name, target_pks)
            if response:
                return response
            else:
                # message in session first
                return ResponseRedirect(".")

        raise ValueError("Invalid action")

    def perform_action(self, action: str, target_pks: list) -> Response | None:
        raise NotImplementedError

    def get_objects(self) -> list:
        return []

    def get_fields(self) -> list:
        return (
            self.fields.copy()
        )  # Avoid mutating the class attribute if using append etc

    def get_actions(self) -> dict[str]:
        return self.actions.copy()  # Avoid mutating the class attribute itself

    def get_displays(self) -> list[str]:
        return self.displays.copy()  # Avoid mutating the class attribute itself

    def get_field_value(self, obj, field: str):
        try:
            # Try basic dict lookup first
            if field in obj:
                return obj[field]
        except TypeError:
            pass

        # Try dot notation
        if "." in field:
            field, subfield = field.split(".", 1)
            return self.get_field_value(obj[field], subfield)

        # Try regular object attribute
        attr = getattr(obj, field)

        # Call if it's callable
        if callable(attr):
            return attr()
        else:
            return attr

    def get_object_pk(self, obj):
        try:
            return self.get_field_value(obj, "pk")
        except AttributeError:
            return self.get_field_value(obj, "id")

    def get_field_value_template(self, obj, field: str, value):
        type_str = type(value).__name__.lower()
        return [
            f"staff/values/{type_str}.html",  # Create a template per-type
            f"staff/values/{field}.html",  # Or for specific field names
            "staff/values/default.html",
        ]

    def get_create_url(self) -> str | None:
        return None

    def get_detail_url(self, obj) -> str | None:
        return None

    def get_update_url(self, obj) -> str | None:
        return None

    def get_delete_url(self, obj) -> str | None:
        return None

    def get_object_url(self, obj) -> str | None:
        if url := self.get_detail_url(obj):
            return url
        if url := self.get_update_url(obj):
            return url
        if url := self.get_delete_url(obj):
            return url
        return None

    def get_object_links(self, obj) -> dict[str]:
        links = {}
        if self.get_detail_url(obj):
            links["Detail"] = self.get_detail_url(obj)
        if self.get_update_url(obj):
            links["Update"] = self.get_update_url(obj)
        if self.get_delete_url(obj):
            links["Delete"] = self.get_delete_url(obj)
        return links

    def get_links(self):
        links = super().get_links()

        # Not tied to a specific object
        if create_url := self.get_create_url():
            links["Create"] = create_url

        return links


class StaffDetailView(StaffView, DetailView):
    template_name = None
    nav_section = ""

    def get_template_context(self):
        context = super().get_template_context()
        context["get_field_value"] = self.get_field_value
        return context

    def get_template_names(self) -> list[str]:
        return super().get_template_names() + [
            "staff/detail.html",
        ]

    def get_description(self):
        return repr(self.object)

    def get_field_value(self, obj, field: str):
        return getattr(obj, field)

    def get_update_url(self, obj) -> str | None:
        return None

    def get_delete_url(self, obj) -> str | None:
        return None

    def get_list_url(self) -> str | None:
        return None

    def get_links(self):
        links = super().get_links()
        if hasattr(self.object, "get_absolute_url"):
            links["View in app"] = self.object.get_absolute_url()
        if update_url := self.get_update_url(self.object):
            links["Update"] = update_url
        if delete_url := self.get_delete_url(self.object):
            links["Delete"] = delete_url
        return links

    def get_success_url(self, form):
        if detail_url := self.get_detail_url(self.object):
            return detail_url

        if list_url := self.get_list_url():
            return list_url

        if update_url := self.get_update_url(self.object):
            return update_url

        return super().get_success_url(form)


class StaffUpdateView(StaffView, UpdateView):
    template_name = None
    nav_section = ""

    def get_template_names(self) -> list[str]:
        return super().get_template_names() + [
            "staff/update.html",
        ]

    def get_detail_url(self, obj) -> str | None:
        return None

    def get_delete_url(self, obj) -> str | None:
        return None

    def get_list_url(self) -> str | None:
        return None

    def get_description(self):
        return repr(self.object)

    def get_links(self):
        links = super().get_links()

        if hasattr(self.object, "get_absolute_url"):
            links["View in app"] = self.object.get_absolute_url()

        if detail_url := self.get_detail_url(self.object):
            links["Detail"] = detail_url

        if delete_url := self.get_delete_url(self.object):
            links["Delete"] = delete_url

        return links


class StaffCreateView(StaffView, CreateView):
    template_name = None

    def get_template_names(self) -> list[str]:
        return super().get_template_names() + [
            "staff/create.html",
        ]

    def get_list_url(self) -> str | None:
        return None


class StaffDeleteView(StaffView, DeleteView):
    template_name = "staff/delete.html"
    nav_section = ""

    def get_detail_url(self, obj) -> str | None:
        return None

    def get_update_url(self, obj) -> str | None:
        return None

    def get_list_url(self) -> str | None:
        return None

    def get_links(self):
        links = super().get_links()

        if hasattr(self.object, "get_absolute_url"):
            links["View in app"] = self.object.get_absolute_url()

        if detail_url := self.get_detail_url(self.object):
            links["Detail"] = detail_url

        if update_url := self.get_update_url(self.object):
            links["Update"] = update_url

        return links

    def get_success_url(self, form):
        if list_url := self.get_list_url():
            return list_url

        return super().get_success_url(form)



from app.users.models import User


def test_staff_login_required(db, client):
    # Login required
    assert client.get("/staff/").status_code == 302

    user = User.objects.create(username="test")
    client.force_login(user)

    # Not staff yet
    assert client.get("/staff/").status_code == 404

    user.is_staff = True
    user.save()

    # Now staff
    assert client.get("/staff/").status_code == 200



import sys
import traceback

from plain.runtime import settings


class Toolbar:
    def __init__(self, request):
        self.request = request
        self.version = settings.TOOLBAR_VERSION
        self.metadata = {
            "Request ID": request.unique_id,
        }

    def should_render(self):
        if settings.DEBUG:
            return True

        if hasattr(self.request, "impersonator"):
            return self.request.impersonator.is_staff

        if self.request.user:
            return self.request.user.is_staff

        return False

    def request_exception(self):
        # We can capture the exception currently being handled here, if any.
        exception = sys.exception()

        if exception:
            exception._traceback_string = "".join(
                traceback.format_tb(exception.__traceback__)
            )

        return exception



from plain.runtime import settings
from plain.templates import register_template_extension, register_template_filter
from plain.templates.jinja.extensions import InclusionTagExtension
from plain.utils.module_loading import import_string

from .views.registry import registry


@register_template_extension
class ToolbarExtension(InclusionTagExtension):
    tags = {"toolbar"}
    template_name = "toolbar/toolbar.html"

    def get_context(self, context, *args, **kwargs):
        if isinstance(settings.TOOLBAR_CLASS, str):
            cls = import_string(settings.TOOLBAR_CLASS)
        else:
            cls = settings.TOOLBAR_CLASS
        context.vars["toolbar"] = cls(request=context["request"])
        return context


@register_template_filter
def get_admin_model_detail_url(obj):
    return registry.get_model_detail_url(obj)
